@@ -8,43 +8,84 @@ API for manipulating HTTP/2 settings, keeping track of both the current active
 state of the settings and the unacknowledged future values of the settings.
 """
 import collections
+import enum
 
 from hyperframe.frame import SettingsFrame
 
-from h2.errors import PROTOCOL_ERROR, FLOW_CONTROL_ERROR
+from h2.errors import ErrorCodes
 from h2.exceptions import InvalidSettingsValueError
 
 
-# Aliases for all the settings values.
+class SettingCodes(enum.IntEnum):
+    """
+    All known HTTP/2 setting codes.
 
-#: Allows the sender to inform the remote endpoint of the maximum size of the
-#: header compression table used to decode header blocks, in octets.
-HEADER_TABLE_SIZE = SettingsFrame.HEADER_TABLE_SIZE
+    .. versionadded:: 2.6.0
+    """
 
-#: This setting can be used to disable server push. To disable server push on
-#: a client, set this to 0.
-ENABLE_PUSH = SettingsFrame.ENABLE_PUSH
+    #: Allows the sender to inform the remote endpoint of the maximum size of
+    #: the header compression table used to decode header blocks, in octets.
+    HEADER_TABLE_SIZE = SettingsFrame.HEADER_TABLE_SIZE
 
-#: Indicates the maximum number of concurrent streams that the sender will
-#: allow.
-MAX_CONCURRENT_STREAMS = SettingsFrame.MAX_CONCURRENT_STREAMS
+    #: This setting can be used to disable server push. To disable server push
+    #: on a client, set this to 0.
+    ENABLE_PUSH = SettingsFrame.ENABLE_PUSH
 
-#: Indicates the sender's initial window size (in octets) for stream-level flow
-#: control.
-INITIAL_WINDOW_SIZE = SettingsFrame.INITIAL_WINDOW_SIZE
+    #: Indicates the maximum number of concurrent streams that the sender will
+    #: allow.
+    MAX_CONCURRENT_STREAMS = SettingsFrame.MAX_CONCURRENT_STREAMS
 
-#: Indicates the size of the largest frame payload that the sender is willing
-#: to receive, in octets.
-try:  # Platform-specific: Hyperframe < 4.0.0
-    MAX_FRAME_SIZE = SettingsFrame.SETTINGS_MAX_FRAME_SIZE
-except AttributeError:  # Platform-specific: Hyperframe >= 4.0.0
+    #: Indicates the sender's initial window size (in octets) for stream-level
+    #: flow control.
+    INITIAL_WINDOW_SIZE = SettingsFrame.INITIAL_WINDOW_SIZE
+
+    #: Indicates the size of the largest frame payload that the sender is
+    #: willing to receive, in octets.
     MAX_FRAME_SIZE = SettingsFrame.MAX_FRAME_SIZE
 
+    #: This advisory setting informs a peer of the maximum size of header list
+    #: that the sender is prepared to accept, in octets.  The value is based on
+    #: the uncompressed size of header fields, including the length of the name
+    #: and value in octets plus an overhead of 32 octets for each header field.
+    MAX_HEADER_LIST_SIZE = SettingsFrame.MAX_HEADER_LIST_SIZE
 
-#: A value structure for storing changed settings.
-ChangedSetting = collections.namedtuple(
-    'ChangedSetting', ['setting', 'original_value', 'new_value']
-)
+
+def _setting_code_from_int(code):
+    """
+    Given an integer setting code, returns either one of :class:`SettingCodes
+    <h2.settings.SettingCodes>` or, if not present in the known set of codes,
+    returns the integer directly.
+    """
+    try:
+        return SettingCodes(code)
+    except ValueError:
+        return code
+
+
+class ChangedSetting:
+
+    def __init__(self, setting, original_value, new_value):
+        #: The setting code given. Either one of :class:`SettingCodes
+        #: <h2.settings.SettingCodes>` or ``int``
+        #:
+        #: .. versionchanged:: 2.6.0
+        self.setting = setting
+
+        #: The original value before being changed.
+        self.original_value = original_value
+
+        #: The new value after being changed.
+        self.new_value = new_value
+
+    def __repr__(self):
+        return (
+            "ChangedSetting(setting=%s, original_value=%s, "
+            "new_value=%s)"
+        ) % (
+            self.setting,
+            self.original_value,
+            self.new_value
+        )
 
 
 class Settings(collections.MutableMapping):
@@ -72,6 +113,9 @@ class Settings(collections.MutableMapping):
     .. versionchanged:: 2.2.0
        Added the ``initial_values`` parameter.
 
+    .. versionchanged:: 2.5.0
+       Added the ``max_header_list_size`` property.
+
     :param client: (optional) Whether these settings should be defaulted for a
         client implementation or a server implementation. Defaults to ``True``.
     :type client: ``bool``
@@ -87,10 +131,10 @@ class Settings(collections.MutableMapping):
         #
         # This contains the default values for HTTP/2.
         self._settings = {
-            HEADER_TABLE_SIZE: collections.deque([4096]),
-            ENABLE_PUSH: collections.deque([int(client)]),
-            INITIAL_WINDOW_SIZE: collections.deque([65535]),
-            MAX_FRAME_SIZE: collections.deque([16384]),
+            SettingCodes.HEADER_TABLE_SIZE: collections.deque([4096]),
+            SettingCodes.ENABLE_PUSH: collections.deque([int(client)]),
+            SettingCodes.INITIAL_WINDOW_SIZE: collections.deque([65535]),
+            SettingCodes.MAX_FRAME_SIZE: collections.deque([16384]),
         }
         if initial_values is not None:
             for key, value in initial_values.items():
@@ -128,61 +172,76 @@ class Settings(collections.MutableMapping):
     def header_table_size(self):
         """
         The current value of the :data:`HEADER_TABLE_SIZE
-        <h2.settings.HEADER_TABLE_SIZE>` setting.
+        <h2.settings.SettingCodes.HEADER_TABLE_SIZE>` setting.
         """
-        return self[HEADER_TABLE_SIZE]
+        return self[SettingCodes.HEADER_TABLE_SIZE]
 
     @header_table_size.setter
     def header_table_size(self, value):
-        self[HEADER_TABLE_SIZE] = value
+        self[SettingCodes.HEADER_TABLE_SIZE] = value
 
     @property
     def enable_push(self):
         """
-        The current value of the :data:`ENABLE_PUSH <h2.settings.ENABLE_PUSH>`
-        setting.
+        The current value of the :data:`ENABLE_PUSH
+        <h2.settings.SettingCodes.ENABLE_PUSH>` setting.
         """
-        return self[ENABLE_PUSH]
+        return self[SettingCodes.ENABLE_PUSH]
 
     @enable_push.setter
     def enable_push(self, value):
-        self[ENABLE_PUSH] = value
+        self[SettingCodes.ENABLE_PUSH] = value
 
     @property
     def initial_window_size(self):
         """
         The current value of the :data:`INITIAL_WINDOW_SIZE
-        <h2.settings.INITIAL_WINDOW_SIZE>` setting.
+        <h2.settings.SettingCodes.INITIAL_WINDOW_SIZE>` setting.
         """
-        return self[INITIAL_WINDOW_SIZE]
+        return self[SettingCodes.INITIAL_WINDOW_SIZE]
 
     @initial_window_size.setter
     def initial_window_size(self, value):
-        self[INITIAL_WINDOW_SIZE] = value
+        self[SettingCodes.INITIAL_WINDOW_SIZE] = value
 
     @property
     def max_frame_size(self):
         """
         The current value of the :data:`MAX_FRAME_SIZE
-        <h2.settings.MAX_FRAME_SIZE>` setting.
+        <h2.settings.SettingCodes.MAX_FRAME_SIZE>` setting.
         """
-        return self[MAX_FRAME_SIZE]
+        return self[SettingCodes.MAX_FRAME_SIZE]
 
     @max_frame_size.setter
     def max_frame_size(self, value):
-        self[MAX_FRAME_SIZE] = value
+        self[SettingCodes.MAX_FRAME_SIZE] = value
 
     @property
     def max_concurrent_streams(self):
         """
         The current value of the :data:`MAX_CONCURRENT_STREAMS
-        <h2.settings.MAX_CONCURRENT_STREAMS>` setting.
+        <h2.settings.SettingCodes.MAX_CONCURRENT_STREAMS>` setting.
         """
-        return self.get(MAX_CONCURRENT_STREAMS, 2**32+1)
+        return self.get(SettingCodes.MAX_CONCURRENT_STREAMS, 2**32+1)
 
     @max_concurrent_streams.setter
     def max_concurrent_streams(self, value):
-        self[MAX_CONCURRENT_STREAMS] = value
+        self[SettingCodes.MAX_CONCURRENT_STREAMS] = value
+
+    @property
+    def max_header_list_size(self):
+        """
+        The current value of the :data:`MAX_HEADER_LIST_SIZE
+        <h2.settings.SettingCodes.MAX_HEADER_LIST_SIZE>` setting. If not set,
+        returns ``None``, which means unlimited.
+
+        .. versionadded:: 2.5.0
+        """
+        return self.get(SettingCodes.MAX_HEADER_LIST_SIZE, None)
+
+    @max_header_list_size.setter
+    def max_header_list_size(self, value):
+        self[SettingCodes.MAX_HEADER_LIST_SIZE] = value
 
     # Implement the MutableMapping API.
     def __getitem__(self, key):
@@ -220,20 +279,35 @@ class Settings(collections.MutableMapping):
     def __len__(self):
         return len(self._settings)
 
+    def __eq__(self, other):
+        if isinstance(other, Settings):
+            return self._settings == other._settings
+        else:
+            return NotImplemented
+
+    def __ne__(self, other):
+        if isinstance(other, Settings):
+            return not self == other
+        else:
+            return NotImplemented
+
 
 def _validate_setting(setting, value):
     """
     Confirms that a specific setting has a well-formed value. If the setting is
     invalid, returns an error code. Otherwise, returns 0 (NO_ERROR).
     """
-    if setting == ENABLE_PUSH:
+    if setting == SettingCodes.ENABLE_PUSH:
         if value not in (0, 1):
-            return PROTOCOL_ERROR
-    elif setting == INITIAL_WINDOW_SIZE:
+            return ErrorCodes.PROTOCOL_ERROR
+    elif setting == SettingCodes.INITIAL_WINDOW_SIZE:
         if not 0 <= value <= 2147483647:  # 2^31 - 1
-            return FLOW_CONTROL_ERROR
-    elif setting == MAX_FRAME_SIZE:
+            return ErrorCodes.FLOW_CONTROL_ERROR
+    elif setting == SettingCodes.MAX_FRAME_SIZE:
         if not 16384 <= value <= 16777215:  # 2^14 and 2^24 - 1
-            return PROTOCOL_ERROR
+            return ErrorCodes.PROTOCOL_ERROR
+    elif setting == SettingCodes.MAX_HEADER_LIST_SIZE:
+        if value < 0:
+            return ErrorCodes.PROTOCOL_ERROR
 
     return 0
